# THIS FILE WAS AUTOMATICALLY GENERATED BY "generate_dicts_from_data_json.py" DO NOT CHANGE MANUALLY!
# ANY CHANGE WILL BE OVERWRITTEN

from typing import Dict, Set

from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId

# from ..ids.buff_id import BuffId
# from ..ids.effect_id import EffectId

UNIT_ABILITIES: Dict[UnitTypeId, Set[AbilityId]] = {
    UnitTypeId.ADEPT: {
        AbilityId.ADEPTPHASESHIFT_ADEPTPHASESHIFT, AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ADEPTPHASESHIFT: {
        AbilityId.ATTACK_ATTACK, AbilityId.CANCEL_ADEPTSHADEPHASESHIFT, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ARBITERMP: {
        AbilityId.ARBITERMPRECALL_ARBITERMPRECALL, AbilityId.ARBITERMPSTASISFIELD_ARBITERMPSTASISFIELD,
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ARCHON: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ARMORY: {
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1, AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3, AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL2,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL3, AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1,
        AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2, AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3
    },
    UnitTypeId.AUTOTURRET: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.BANELING: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_BUILDINGATTACKON, AbilityId.BURROWDOWN_BANELING,
        AbilityId.EXPLODE_EXPLODE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.BANELINGBURROWED: {AbilityId.BURROWUP_BANELING, AbilityId.EXPLODE_EXPLODE},
    UnitTypeId.BANELINGCOCOON: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.BANELINGNEST: {AbilityId.RESEARCH_CENTRIFUGALHOOKS},
    UnitTypeId.BANSHEE: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_CLOAKON_BANSHEE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.BARRACKS: {
        AbilityId.BARRACKSTRAIN_GHOST, AbilityId.BARRACKSTRAIN_MARAUDER, AbilityId.BARRACKSTRAIN_MARINE,
        AbilityId.BARRACKSTRAIN_REAPER, AbilityId.BUILD_REACTOR_BARRACKS, AbilityId.BUILD_TECHLAB_BARRACKS,
        AbilityId.LIFT_BARRACKS, AbilityId.RALLY_BUILDING, AbilityId.SMART
    },
    UnitTypeId.BARRACKSFLYING: {
        AbilityId.BUILD_REACTOR_BARRACKS, AbilityId.BUILD_TECHLAB_BARRACKS, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_BARRACKS, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.BARRACKSTECHLAB:
    {AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK, AbilityId.RESEARCH_COMBATSHIELD, AbilityId.RESEARCH_CONCUSSIVESHELLS},
    UnitTypeId.BATTLECRUISER: {
        AbilityId.ATTACK_BATTLECRUISER, AbilityId.EFFECT_TACTICALJUMP, AbilityId.HOLDPOSITION_BATTLECRUISER,
        AbilityId.MOVE_BATTLECRUISER, AbilityId.PATROL_BATTLECRUISER, AbilityId.SMART, AbilityId.STOP_BATTLECRUISER,
        AbilityId.YAMATO_YAMATOGUN
    },
    UnitTypeId.BROODLING: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.BROODLORD: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.BUNKER: {AbilityId.EFFECT_SALVAGE, AbilityId.LOAD_BUNKER, AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.BYPASSARMORDRONE: {AbilityId.ATTACK_ATTACK, AbilityId.MOVE_MOVE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.CARRIER: {
        AbilityId.ATTACK_ATTACK, AbilityId.BUILD_INTERCEPTORS, AbilityId.CANCEL_HANGARQUEUE5,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELING: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELINGMARINE: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELINGMARINESHIELD: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELINGZEALOT: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELINGZERGLING: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CHANGELINGZERGLINGWINGS: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.COLOSSUS: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.COMMANDCENTER: {
        AbilityId.COMMANDCENTERTRAIN_SCV, AbilityId.LIFT_COMMANDCENTER, AbilityId.LOADALL_COMMANDCENTER,
        AbilityId.RALLY_COMMANDCENTER, AbilityId.SMART, AbilityId.UPGRADETOORBITAL_ORBITALCOMMAND,
        AbilityId.UPGRADETOPLANETARYFORTRESS_PLANETARYFORTRESS
    },
    UnitTypeId.COMMANDCENTERFLYING: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.LAND_COMMANDCENTER, AbilityId.LOADALL_COMMANDCENTER, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.CORRUPTOR: {
        AbilityId.ATTACK_ATTACK, AbilityId.CAUSTICSPRAY_CAUSTICSPRAY, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHTOBROODLORD_BROODLORD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.CORSAIRMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.CORSAIRMPDISRUPTIONWEB_CORSAIRMPDISRUPTIONWEB, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.CREEPTUMOR: {AbilityId.BUILD_CREEPTUMOR_TUMOR, AbilityId.SMART},
    UnitTypeId.CREEPTUMORBURROWED: {AbilityId.BUILD_CREEPTUMOR, AbilityId.BUILD_CREEPTUMOR_TUMOR, AbilityId.SMART},
    UnitTypeId.CREEPTUMORQUEEN: {AbilityId.BUILD_CREEPTUMOR_TUMOR, AbilityId.SMART},
    UnitTypeId.CYBERNETICSCORE: {
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL1,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL2,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL3,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2,
        AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3, AbilityId.RESEARCH_WARPGATE
    },
    UnitTypeId.CYCLONE: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.LOCKON_LOCKON, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.DARKSHRINE: {AbilityId.RESEARCH_SHADOWSTRIKE},
    UnitTypeId.DARKTEMPLAR: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_SHADOWSTRIDE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_ARCHON,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.DEFILERMP: {
        AbilityId.DEFILERMPBURROW_BURROWDOWN, AbilityId.DEFILERMPCONSUME_DEFILERMPCONSUME,
        AbilityId.DEFILERMPDARKSWARM_DEFILERMPDARKSWARM, AbilityId.DEFILERMPPLAGUE_DEFILERMPPLAGUE,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.DEFILERMPBURROWED: {AbilityId.DEFILERMPUNBURROW_BURROWUP},
    UnitTypeId.DEVOURERMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.DISRUPTOR: {
        AbilityId.EFFECT_PURIFICATIONNOVA, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.DISRUPTORPHASED: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.DRONE: {
        AbilityId.ATTACK_ATTACK, AbilityId.BUILD_LURKERDEN, AbilityId.BURROWDOWN_DRONE, AbilityId.EFFECT_SPRAY_ZERG,
        AbilityId.HARVEST_GATHER_DRONE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.ZERGBUILD_BANELINGNEST, AbilityId.ZERGBUILD_EVOLUTIONCHAMBER,
        AbilityId.ZERGBUILD_EXTRACTOR, AbilityId.ZERGBUILD_HATCHERY, AbilityId.ZERGBUILD_HYDRALISKDEN,
        AbilityId.ZERGBUILD_INFESTATIONPIT, AbilityId.ZERGBUILD_NYDUSNETWORK, AbilityId.ZERGBUILD_ROACHWARREN,
        AbilityId.ZERGBUILD_SPAWNINGPOOL, AbilityId.ZERGBUILD_SPINECRAWLER, AbilityId.ZERGBUILD_SPIRE,
        AbilityId.ZERGBUILD_SPORECRAWLER, AbilityId.ZERGBUILD_ULTRALISKCAVERN
    },
    UnitTypeId.DRONEBURROWED: {AbilityId.BURROWUP_DRONE},
    UnitTypeId.EGG: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.ELSECARO_COLONIST_HUT: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.ENGINEERINGBAY: {
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2,
        AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3, AbilityId.RESEARCH_HISECAUTOTRACKING,
        AbilityId.RESEARCH_TERRANSTRUCTUREARMORUPGRADE
    },
    UnitTypeId.EVOLUTIONCHAMBER: {
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1, AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2,
        AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3, AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1,
        AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2, AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1, AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2,
        AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3
    },
    UnitTypeId.FACTORY: {
        AbilityId.BUILD_REACTOR_FACTORY, AbilityId.BUILD_TECHLAB_FACTORY, AbilityId.FACTORYTRAIN_HELLION,
        AbilityId.FACTORYTRAIN_SIEGETANK, AbilityId.FACTORYTRAIN_THOR, AbilityId.FACTORYTRAIN_WIDOWMINE,
        AbilityId.LIFT_FACTORY, AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.TRAIN_CYCLONE,
        AbilityId.TRAIN_HELLBAT
    },
    UnitTypeId.FACTORYFLYING: {
        AbilityId.BUILD_REACTOR_FACTORY, AbilityId.BUILD_TECHLAB_FACTORY, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_FACTORY, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.FACTORYTECHLAB: {
        AbilityId.RESEARCH_CYCLONELOCKONDAMAGE, AbilityId.RESEARCH_DRILLINGCLAWS, AbilityId.RESEARCH_INFERNALPREIGNITER,
        AbilityId.RESEARCH_SMARTSERVOS
    },
    UnitTypeId.FLEETBEACON: {
        AbilityId.FLEETBEACONRESEARCH_RESEARCHVOIDRAYSPEEDUPGRADE,
        AbilityId.FLEETBEACONRESEARCH_TEMPESTRESEARCHGROUNDATTACKUPGRADE, AbilityId.RESEARCH_PHOENIXANIONPULSECRYSTALS
    },
    UnitTypeId.FORGE: {
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1, AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3, AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1,
        AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2, AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1, AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2,
        AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3
    },
    UnitTypeId.FUSIONCORE: {
        AbilityId.FUSIONCORERESEARCH_RESEARCHBALLISTICRANGE, AbilityId.FUSIONCORERESEARCH_RESEARCHRAPIDREIGNITIONSYSTEM,
        AbilityId.RESEARCH_BATTLECRUISERWEAPONREFIT
    },
    UnitTypeId.GATEWAY: {
        AbilityId.GATEWAYTRAIN_DARKTEMPLAR, AbilityId.GATEWAYTRAIN_HIGHTEMPLAR, AbilityId.GATEWAYTRAIN_SENTRY,
        AbilityId.GATEWAYTRAIN_STALKER, AbilityId.GATEWAYTRAIN_ZEALOT, AbilityId.MORPH_WARPGATE,
        AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.TRAIN_ADEPT
    },
    UnitTypeId.GHOST: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_CLOAKON_GHOST, AbilityId.BEHAVIOR_HOLDFIREON_GHOST,
        AbilityId.EFFECT_GHOSTSNIPE, AbilityId.EMP_EMP, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.GHOSTACADEMY: {
        AbilityId.BUILD_NUKE, AbilityId.GHOSTACADEMYRESEARCH_RESEARCHENHANCEDSHOCKWAVES,
        AbilityId.RESEARCH_PERSONALCLOAKING
    },
    UnitTypeId.GHOSTNOVA: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_CLOAKON_GHOST, AbilityId.BEHAVIOR_HOLDFIREON_GHOST,
        AbilityId.EFFECT_GHOSTSNIPE, AbilityId.EMP_EMP, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.GREATERSPIRE: {
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1, AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3, AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2, AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3
    },
    UnitTypeId.GUARDIANMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HATCHERY: {
        AbilityId.RALLY_HATCHERY_UNITS, AbilityId.RALLY_HATCHERY_WORKERS, AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE, AbilityId.SMART, AbilityId.TRAINQUEEN_QUEEN,
        AbilityId.UPGRADETOLAIR_LAIR
    },
    UnitTypeId.HELLION: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_HELLBAT, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HELLIONTANK: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_HELLION, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HERC: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HERCPLACEMENT: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HIGHTEMPLAR: {
        AbilityId.ATTACK_ATTACK, AbilityId.FEEDBACK_FEEDBACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_ARCHON,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.PSISTORM_PSISTORM, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.HIVE: {
        AbilityId.RALLY_HATCHERY_UNITS, AbilityId.RALLY_HATCHERY_WORKERS, AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE, AbilityId.SMART, AbilityId.TRAINQUEEN_QUEEN
    },
    UnitTypeId.HYDRALISK: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_HYDRALISK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_LURKER,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.HYDRALISKBURROWED: {AbilityId.BURROWUP_HYDRALISK},
    UnitTypeId.HYDRALISKDEN: {AbilityId.RESEARCH_GROOVEDSPINES, AbilityId.RESEARCH_MUSCULARAUGMENTS},
    UnitTypeId.IMMORTAL: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.INFESTATIONPIT: {AbilityId.RESEARCH_NEURALPARASITE, AbilityId.RESEARCH_PATHOGENGLANDS},
    UnitTypeId.INFESTEDTERRANSEGG:
    {AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART},
    UnitTypeId.INFESTOR: {
        AbilityId.AMORPHOUSARMORCLOUD_AMORPHOUSARMORCLOUD, AbilityId.BURROWDOWN_INFESTOR,
        AbilityId.BURROWDOWN_INFESTORTERRAN, AbilityId.FUNGALGROWTH_FUNGALGROWTH, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.NEURALPARASITE_NEURALPARASITE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.INFESTORBURROWED: {
        AbilityId.BURROWUP_INFESTOR, AbilityId.BURROWUP_INFESTORTERRAN, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.NEURALPARASITE_NEURALPARASITE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.INFESTORTERRAN: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_INFESTORTERRAN, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.INFESTORTERRANBURROWED: {AbilityId.BURROWUP_INFESTORTERRAN},
    UnitTypeId.INTERCEPTOR: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.LAIR: {
        AbilityId.RALLY_HATCHERY_UNITS, AbilityId.RALLY_HATCHERY_WORKERS, AbilityId.RESEARCH_BURROW,
        AbilityId.RESEARCH_PNEUMATIZEDCARAPACE, AbilityId.SMART, AbilityId.TRAINQUEEN_QUEEN,
        AbilityId.UPGRADETOHIVE_HIVE
    },
    UnitTypeId.LARVA: {
        AbilityId.LARVATRAIN_CORRUPTOR, AbilityId.LARVATRAIN_DRONE, AbilityId.LARVATRAIN_HYDRALISK,
        AbilityId.LARVATRAIN_INFESTOR, AbilityId.LARVATRAIN_MUTALISK, AbilityId.LARVATRAIN_OVERLORD,
        AbilityId.LARVATRAIN_ROACH, AbilityId.LARVATRAIN_ULTRALISK, AbilityId.LARVATRAIN_VIPER,
        AbilityId.LARVATRAIN_ZERGLING, AbilityId.TRAIN_SWARMHOST
    },
    UnitTypeId.LIBERATOR: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_LIBERATORAGMODE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.LIBERATORAG:
    {AbilityId.ATTACK_ATTACK, AbilityId.MORPH_LIBERATORAAMODE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.LOCUSTMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.LOCUSTMPFLYING: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_LOCUSTSWOOP, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.LURKERDENMP: {AbilityId.LURKERDENRESEARCH_RESEARCHLURKERRANGE, AbilityId.RESEARCH_ADAPTIVETALONS},
    UnitTypeId.LURKERMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_LURKER, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.LURKERMPBURROWED: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_HOLDFIREON_LURKER, AbilityId.BURROWUP_LURKER, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.LURKERMPEGG: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.MARAUDER: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_STIM_MARAUDER, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.MARINE: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_STIM_MARINE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.MEDIVAC: {
        AbilityId.EFFECT_MEDIVACIGNITEAFTERBURNERS, AbilityId.HOLDPOSITION_HOLD, AbilityId.LOAD_MEDIVAC,
        AbilityId.MEDIVACHEAL_HEAL, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.MISSILETURRET: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.MOTHERSHIP: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_MASSRECALL_STRATEGICRECALL, AbilityId.EFFECT_TIMEWARP,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.MOTHERSHIPCORE: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_MASSRECALL_MOTHERSHIPCORE, AbilityId.EFFECT_PHOTONOVERCHARGE,
        AbilityId.EFFECT_TIMEWARP, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_MOTHERSHIP, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.MULE: {
        AbilityId.EFFECT_REPAIR_MULE, AbilityId.HARVEST_GATHER_MULE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.MUTALISK: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.NEXUS: {
        AbilityId.BATTERYOVERCHARGE_BATTERYOVERCHARGE, AbilityId.EFFECT_CHRONOBOOSTENERGYCOST,
        AbilityId.EFFECT_MASSRECALL_NEXUS, AbilityId.NEXUSTRAINMOTHERSHIP_MOTHERSHIP, AbilityId.NEXUSTRAIN_PROBE,
        AbilityId.RALLY_NEXUS, AbilityId.SMART
    },
    UnitTypeId.NYDUSCANAL: {AbilityId.LOAD_NYDUSWORM, AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.NYDUSCANALATTACKER: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.NYDUSCANALCREEPER:
    {AbilityId.ATTACK_ATTACK, AbilityId.DIGESTERCREEPSPRAY_DIGESTERCREEPSPRAY, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.NYDUSNETWORK: {
        AbilityId.BUILD_NYDUSWORM, AbilityId.LOAD_NYDUSNETWORK, AbilityId.RALLY_BUILDING, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.OBSERVER: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_SURVEILLANCEMODE, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.OBSERVERSIEGEMODE: {AbilityId.MORPH_OBSERVERMODE, AbilityId.STOP_STOP},
    UnitTypeId.ORACLE: {
        AbilityId.ATTACK_ATTACK, AbilityId.BEHAVIOR_PULSARBEAMON, AbilityId.BUILD_STASISTRAP,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.ORACLEREVELATION_ORACLEREVELATION,
        AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ORBITALCOMMAND: {
        AbilityId.CALLDOWNMULE_CALLDOWNMULE, AbilityId.COMMANDCENTERTRAIN_SCV, AbilityId.LIFT_ORBITALCOMMAND,
        AbilityId.RALLY_COMMANDCENTER, AbilityId.SCANNERSWEEP_SCAN, AbilityId.SMART, AbilityId.SUPPLYDROP_SUPPLYDROP
    },
    UnitTypeId.ORBITALCOMMANDFLYING: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.LAND_ORBITALCOMMAND, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.OVERLORD: {
        AbilityId.BEHAVIOR_GENERATECREEPON, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_OVERLORDTRANSPORT,
        AbilityId.MORPH_OVERSEER, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.OVERLORDTRANSPORT: {
        AbilityId.BEHAVIOR_GENERATECREEPON, AbilityId.HOLDPOSITION_HOLD, AbilityId.LOAD_OVERLORD,
        AbilityId.MORPH_OVERSEER, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.OVERSEER: {
        AbilityId.CONTAMINATE_CONTAMINATE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_OVERSIGHTMODE,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.SPAWNCHANGELING_SPAWNCHANGELING, AbilityId.STOP_STOP
    },
    UnitTypeId.OVERSEERSIEGEMODE: {
        AbilityId.CONTAMINATE_CONTAMINATE, AbilityId.MORPH_OVERSEERMODE, AbilityId.SMART,
        AbilityId.SPAWNCHANGELING_SPAWNCHANGELING, AbilityId.STOP_STOP
    },
    UnitTypeId.PHOENIX: {
        AbilityId.ATTACK_ATTACK, AbilityId.GRAVITONBEAM_GRAVITONBEAM, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.PHOTONCANNON: {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.PLANETARYFORTRESS: {
        AbilityId.ATTACK_ATTACK, AbilityId.COMMANDCENTERTRAIN_SCV, AbilityId.LOADALL_COMMANDCENTER,
        AbilityId.RALLY_COMMANDCENTER, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.PROBE: {
        AbilityId.ATTACK_ATTACK, AbilityId.BUILD_SHIELDBATTERY, AbilityId.EFFECT_SPRAY_PROTOSS,
        AbilityId.HARVEST_GATHER_PROBE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.PROTOSSBUILD_ASSIMILATOR, AbilityId.PROTOSSBUILD_CYBERNETICSCORE, AbilityId.PROTOSSBUILD_DARKSHRINE,
        AbilityId.PROTOSSBUILD_FLEETBEACON, AbilityId.PROTOSSBUILD_FORGE, AbilityId.PROTOSSBUILD_GATEWAY,
        AbilityId.PROTOSSBUILD_NEXUS, AbilityId.PROTOSSBUILD_PHOTONCANNON, AbilityId.PROTOSSBUILD_PYLON,
        AbilityId.PROTOSSBUILD_ROBOTICSBAY, AbilityId.PROTOSSBUILD_ROBOTICSFACILITY, AbilityId.PROTOSSBUILD_STARGATE,
        AbilityId.PROTOSSBUILD_TEMPLARARCHIVE, AbilityId.PROTOSSBUILD_TWILIGHTCOUNCIL, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.QUEEN: {
        AbilityId.ATTACK_ATTACK, AbilityId.BUILD_CREEPTUMOR, AbilityId.BUILD_CREEPTUMOR_QUEEN,
        AbilityId.BURROWDOWN_QUEEN, AbilityId.EFFECT_INJECTLARVA, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.TRANSFUSION_TRANSFUSION
    },
    UnitTypeId.QUEENBURROWED: {AbilityId.BURROWUP_QUEEN},
    UnitTypeId.QUEENMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.QUEENMPENSNARE_QUEENMPENSNARE, AbilityId.QUEENMPINFESTCOMMANDCENTER_QUEENMPINFESTCOMMANDCENTER,
        AbilityId.QUEENMPSPAWNBROODLINGS_QUEENMPSPAWNBROODLINGS, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.RAVAGER: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_RAVAGER, AbilityId.EFFECT_CORROSIVEBILE,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.RAVAGERBURROWED: {AbilityId.BURROWUP_RAVAGER},
    UnitTypeId.RAVAGERCOCOON: {AbilityId.RALLY_BUILDING, AbilityId.SMART},
    UnitTypeId.RAVEN: {
        AbilityId.BUILDAUTOTURRET_AUTOTURRET, AbilityId.EFFECT_ANTIARMORMISSILE, AbilityId.EFFECT_INTERFERENCEMATRIX,
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.RAVENREPAIRDRONE: {AbilityId.EFFECT_REPAIR_REPAIRDRONE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.REAPER: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.KD8CHARGE_KD8CHARGE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.REPLICANT: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ROACH: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_ROACH, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHTORAVAGER_RAVAGER, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.ROACHBURROWED: {
        AbilityId.BURROWUP_ROACH, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ROACHWARREN: {AbilityId.RESEARCH_GLIALREGENERATION, AbilityId.RESEARCH_TUNNELINGCLAWS},
    UnitTypeId.ROBOTICSBAY:
    {AbilityId.RESEARCH_EXTENDEDTHERMALLANCE, AbilityId.RESEARCH_GRAVITICBOOSTER, AbilityId.RESEARCH_GRAVITICDRIVE},
    UnitTypeId.ROBOTICSFACILITY: {
        AbilityId.RALLY_BUILDING, AbilityId.ROBOTICSFACILITYTRAIN_COLOSSUS, AbilityId.ROBOTICSFACILITYTRAIN_IMMORTAL,
        AbilityId.ROBOTICSFACILITYTRAIN_OBSERVER, AbilityId.ROBOTICSFACILITYTRAIN_WARPPRISM, AbilityId.SMART,
        AbilityId.TRAIN_DISRUPTOR
    },
    UnitTypeId.SCOURGEMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.SCOUTMP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.SCV: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_REPAIR_SCV, AbilityId.EFFECT_SPRAY_TERRAN,
        AbilityId.HARVEST_GATHER_SCV, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.TERRANBUILD_ARMORY, AbilityId.TERRANBUILD_BARRACKS,
        AbilityId.TERRANBUILD_BUNKER, AbilityId.TERRANBUILD_COMMANDCENTER, AbilityId.TERRANBUILD_ENGINEERINGBAY,
        AbilityId.TERRANBUILD_FACTORY, AbilityId.TERRANBUILD_FUSIONCORE, AbilityId.TERRANBUILD_GHOSTACADEMY,
        AbilityId.TERRANBUILD_MISSILETURRET, AbilityId.TERRANBUILD_REFINERY, AbilityId.TERRANBUILD_SENSORTOWER,
        AbilityId.TERRANBUILD_STARPORT, AbilityId.TERRANBUILD_SUPPLYDEPOT
    },
    UnitTypeId.SENTRY: {
        AbilityId.ATTACK_ATTACK, AbilityId.FORCEFIELD_FORCEFIELD, AbilityId.GUARDIANSHIELD_GUARDIANSHIELD,
        AbilityId.HALLUCINATION_ADEPT, AbilityId.HALLUCINATION_ARCHON, AbilityId.HALLUCINATION_COLOSSUS,
        AbilityId.HALLUCINATION_DISRUPTOR, AbilityId.HALLUCINATION_HIGHTEMPLAR, AbilityId.HALLUCINATION_IMMORTAL,
        AbilityId.HALLUCINATION_ORACLE, AbilityId.HALLUCINATION_PHOENIX, AbilityId.HALLUCINATION_PROBE,
        AbilityId.HALLUCINATION_STALKER, AbilityId.HALLUCINATION_VOIDRAY, AbilityId.HALLUCINATION_WARPPRISM,
        AbilityId.HALLUCINATION_ZEALOT, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.SIEGETANK: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SIEGEMODE_SIEGEMODE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.SIEGETANKSIEGED:
    {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.UNSIEGE_UNSIEGE},
    UnitTypeId.SPAWNINGPOOL: {AbilityId.RESEARCH_ZERGLINGADRENALGLANDS, AbilityId.RESEARCH_ZERGLINGMETABOLICBOOST},
    UnitTypeId.SPINECRAWLER:
    {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.SPINECRAWLERUPROOT_SPINECRAWLERUPROOT, AbilityId.STOP_STOP},
    UnitTypeId.SPINECRAWLERUPROOTED: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.SPINECRAWLERROOT_SPINECRAWLERROOT, AbilityId.STOP_STOP
    },
    UnitTypeId.SPIRE: {
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1, AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
        AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3, AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1,
        AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2, AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
        AbilityId.UPGRADETOGREATERSPIRE_GREATERSPIRE
    },
    UnitTypeId.SPORECRAWLER:
    {AbilityId.ATTACK_ATTACK, AbilityId.SMART, AbilityId.SPORECRAWLERUPROOT_SPORECRAWLERUPROOT, AbilityId.STOP_STOP},
    UnitTypeId.SPORECRAWLERUPROOTED: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART,
        AbilityId.SPORECRAWLERROOT_SPORECRAWLERROOT, AbilityId.STOP_STOP
    },
    UnitTypeId.STALKER: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_BLINK_STALKER, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.STARGATE: {
        AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.STARGATETRAIN_CARRIER, AbilityId.STARGATETRAIN_ORACLE,
        AbilityId.STARGATETRAIN_PHOENIX, AbilityId.STARGATETRAIN_TEMPEST, AbilityId.STARGATETRAIN_VOIDRAY
    },
    UnitTypeId.STARPORT: {
        AbilityId.BUILD_REACTOR_STARPORT, AbilityId.BUILD_TECHLAB_STARPORT, AbilityId.LIFT_STARPORT,
        AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.STARPORTTRAIN_BANSHEE,
        AbilityId.STARPORTTRAIN_BATTLECRUISER, AbilityId.STARPORTTRAIN_LIBERATOR, AbilityId.STARPORTTRAIN_MEDIVAC,
        AbilityId.STARPORTTRAIN_RAVEN, AbilityId.STARPORTTRAIN_VIKINGFIGHTER
    },
    UnitTypeId.STARPORTFLYING: {
        AbilityId.BUILD_REACTOR_STARPORT, AbilityId.BUILD_TECHLAB_STARPORT, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.LAND_STARPORT, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.STARPORTTECHLAB: {
        AbilityId.RESEARCH_BANSHEECLOAKINGFIELD, AbilityId.RESEARCH_BANSHEEHYPERFLIGHTROTORS,
        AbilityId.RESEARCH_RAVENCORVIDREACTOR
    },
    UnitTypeId.SUPPLYDEPOT: {AbilityId.MORPH_SUPPLYDEPOT_LOWER},
    UnitTypeId.SUPPLYDEPOTLOWERED: {AbilityId.MORPH_SUPPLYDEPOT_RAISE},
    UnitTypeId.SWARMHOSTBURROWEDMP: {AbilityId.EFFECT_SPAWNLOCUSTS, AbilityId.SMART},
    UnitTypeId.SWARMHOSTMP: {
        AbilityId.BURROWDOWN_SWARMHOST, AbilityId.EFFECT_SPAWNLOCUSTS, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.TECHLAB: {
        AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK, AbilityId.RESEARCH_BANSHEECLOAKINGFIELD,
        AbilityId.RESEARCH_COMBATSHIELD, AbilityId.RESEARCH_CONCUSSIVESHELLS, AbilityId.RESEARCH_DRILLINGCLAWS,
        AbilityId.RESEARCH_INFERNALPREIGNITER, AbilityId.RESEARCH_RAVENCORVIDREACTOR
    },
    UnitTypeId.TEMPEST: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.TEMPLARARCHIVE: {AbilityId.RESEARCH_PSISTORM},
    UnitTypeId.THOR: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_THORHIGHIMPACTMODE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.THORAP: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_THOREXPLOSIVEMODE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.TWILIGHTCOUNCIL:
    {AbilityId.RESEARCH_ADEPTRESONATINGGLAIVES, AbilityId.RESEARCH_BLINK, AbilityId.RESEARCH_CHARGE},
    UnitTypeId.ULTRALISK: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_ULTRALISK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ULTRALISKBURROWED: {AbilityId.BURROWUP_ULTRALISK},
    UnitTypeId.ULTRALISKCAVERN: {AbilityId.RESEARCH_ANABOLICSYNTHESIS, AbilityId.RESEARCH_CHITINOUSPLATING},
    UnitTypeId.VIKINGASSAULT: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_VIKINGFIGHTERMODE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.VIKINGFIGHTER: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MORPH_VIKINGASSAULTMODE, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.VIPER: {
        AbilityId.BLINDINGCLOUD_BLINDINGCLOUD, AbilityId.EFFECT_ABDUCT, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.PARASITICBOMB_PARASITICBOMB, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE,
        AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.VIPERCONSUMESTRUCTURE_VIPERCONSUME
    },
    UnitTypeId.VOIDMPIMMORTALREVIVECORPSE:
    {AbilityId.RALLY_BUILDING, AbilityId.SMART, AbilityId.VOIDMPIMMORTALREVIVEREBUILD_IMMORTAL},
    UnitTypeId.VOIDRAY: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_VOIDRAYPRISMATICALIGNMENT, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.WARHOUND: {
        AbilityId.ATTACK_ATTACK, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SMART, AbilityId.STOP_STOP, AbilityId.TORNADOMISSILE_TORNADOMISSILE
    },
    UnitTypeId.WARPGATE: {
        AbilityId.MORPH_GATEWAY, AbilityId.SMART, AbilityId.TRAINWARP_ADEPT, AbilityId.WARPGATETRAIN_DARKTEMPLAR,
        AbilityId.WARPGATETRAIN_HIGHTEMPLAR, AbilityId.WARPGATETRAIN_SENTRY, AbilityId.WARPGATETRAIN_STALKER,
        AbilityId.WARPGATETRAIN_ZEALOT
    },
    UnitTypeId.WARPPRISM: {
        AbilityId.HOLDPOSITION_HOLD, AbilityId.LOAD_WARPPRISM, AbilityId.MORPH_WARPPRISMPHASINGMODE,
        AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.WARPPRISMPHASING:
    {AbilityId.LOAD_WARPPRISM, AbilityId.MORPH_WARPPRISMTRANSPORTMODE, AbilityId.SMART, AbilityId.STOP_STOP},
    UnitTypeId.WIDOWMINE: {
        AbilityId.BURROWDOWN_WIDOWMINE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL,
        AbilityId.SCAN_MOVE, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.WIDOWMINEBURROWED:
    {AbilityId.BURROWUP_WIDOWMINE, AbilityId.SMART, AbilityId.WIDOWMINEATTACK_WIDOWMINEATTACK},
    UnitTypeId.ZEALOT: {
        AbilityId.ATTACK_ATTACK, AbilityId.EFFECT_CHARGE, AbilityId.HOLDPOSITION_HOLD, AbilityId.MOVE_MOVE,
        AbilityId.PATROL_PATROL, AbilityId.SMART, AbilityId.STOP_STOP
    },
    UnitTypeId.ZERGLING: {
        AbilityId.ATTACK_ATTACK, AbilityId.BURROWDOWN_ZERGLING, AbilityId.HOLDPOSITION_HOLD,
        AbilityId.MORPHZERGLINGTOBANELING_BANELING, AbilityId.MOVE_MOVE, AbilityId.PATROL_PATROL, AbilityId.SMART,
        AbilityId.STOP_STOP
    },
    UnitTypeId.ZERGLINGBURROWED: {AbilityId.BURROWUP_ZERGLING}
}



# THIS FILE WAS AUTOMATICALLY GENERATED BY "generate_dicts_from_data_json.py" DO NOT CHANGE MANUALLY!
# ANY CHANGE WILL BE OVERWRITTEN

from typing import Dict, Union

from sc2.ids.ability_id import AbilityId
from sc2.ids.unit_typeid import UnitTypeId
from sc2.ids.upgrade_id import UpgradeId

# from ..ids.buff_id import BuffId
# from ..ids.effect_id import EffectId

RESEARCH_INFO: Dict[UnitTypeId, Dict[UpgradeId, Dict[str, Union[AbilityId, bool, UnitTypeId, UpgradeId]]]] = {
    UnitTypeId.ARMORY: {
        UpgradeId.TERRANSHIPWEAPONSLEVEL1: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1
        },
        UpgradeId.TERRANSHIPWEAPONSLEVEL2: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2,
            'required_upgrade': UpgradeId.TERRANSHIPWEAPONSLEVEL1
        },
        UpgradeId.TERRANSHIPWEAPONSLEVEL3: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3,
            'required_upgrade': UpgradeId.TERRANSHIPWEAPONSLEVEL2
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL1: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL1
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL2: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL2,
            'required_upgrade': UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL1
        },
        UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL3: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL3,
            'required_upgrade': UpgradeId.TERRANVEHICLEANDSHIPARMORSLEVEL2
        },
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL1: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1
        },
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL2: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2,
            'required_upgrade': UpgradeId.TERRANVEHICLEWEAPONSLEVEL1
        },
        UpgradeId.TERRANVEHICLEWEAPONSLEVEL3: {
            'ability': AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3,
            'required_upgrade': UpgradeId.TERRANVEHICLEWEAPONSLEVEL2
        }
    },
    UnitTypeId.BANELINGNEST: {
        UpgradeId.CENTRIFICALHOOKS: {
            'ability': AbilityId.RESEARCH_CENTRIFUGALHOOKS,
            'required_building': UnitTypeId.LAIR
        }
    },
    UnitTypeId.BARRACKSTECHLAB: {
        UpgradeId.PUNISHERGRENADES: {
            'ability': AbilityId.RESEARCH_CONCUSSIVESHELLS
        },
        UpgradeId.SHIELDWALL: {
            'ability': AbilityId.RESEARCH_COMBATSHIELD
        },
        UpgradeId.STIMPACK: {
            'ability': AbilityId.BARRACKSTECHLABRESEARCH_STIMPACK
        }
    },
    UnitTypeId.CYBERNETICSCORE: {
        UpgradeId.PROTOSSAIRARMORSLEVEL1: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSAIRARMORSLEVEL2: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL2,
            'required_building': UnitTypeId.FLEETBEACON,
            'required_upgrade': UpgradeId.PROTOSSAIRARMORSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSAIRARMORSLEVEL3: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL3,
            'required_building': UnitTypeId.FLEETBEACON,
            'required_upgrade': UpgradeId.PROTOSSAIRARMORSLEVEL2,
            'requires_power': True
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL1: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL2: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2,
            'required_building': UnitTypeId.FLEETBEACON,
            'required_upgrade': UpgradeId.PROTOSSAIRWEAPONSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSAIRWEAPONSLEVEL3: {
            'ability': AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3,
            'required_building': UnitTypeId.FLEETBEACON,
            'required_upgrade': UpgradeId.PROTOSSAIRWEAPONSLEVEL2,
            'requires_power': True
        },
        UpgradeId.WARPGATERESEARCH: {
            'ability': AbilityId.RESEARCH_WARPGATE,
            'requires_power': True
        }
    },
    UnitTypeId.DARKSHRINE: {
        UpgradeId.DARKTEMPLARBLINKUPGRADE: {
            'ability': AbilityId.RESEARCH_SHADOWSTRIKE,
            'requires_power': True
        }
    },
    UnitTypeId.ENGINEERINGBAY: {
        UpgradeId.HISECAUTOTRACKING: {
            'ability': AbilityId.RESEARCH_HISECAUTOTRACKING
        },
        UpgradeId.TERRANBUILDINGARMOR: {
            'ability': AbilityId.RESEARCH_TERRANSTRUCTUREARMORUPGRADE
        },
        UpgradeId.TERRANINFANTRYARMORSLEVEL1: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1
        },
        UpgradeId.TERRANINFANTRYARMORSLEVEL2: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2,
            'required_building': UnitTypeId.ARMORY,
            'required_upgrade': UpgradeId.TERRANINFANTRYARMORSLEVEL1
        },
        UpgradeId.TERRANINFANTRYARMORSLEVEL3: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3,
            'required_building': UnitTypeId.ARMORY,
            'required_upgrade': UpgradeId.TERRANINFANTRYARMORSLEVEL2
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL1: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL2: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2,
            'required_building': UnitTypeId.ARMORY,
            'required_upgrade': UpgradeId.TERRANINFANTRYWEAPONSLEVEL1
        },
        UpgradeId.TERRANINFANTRYWEAPONSLEVEL3: {
            'ability': AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3,
            'required_building': UnitTypeId.ARMORY,
            'required_upgrade': UpgradeId.TERRANINFANTRYWEAPONSLEVEL2
        }
    },
    UnitTypeId.EVOLUTIONCHAMBER: {
        UpgradeId.ZERGGROUNDARMORSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1
        },
        UpgradeId.ZERGGROUNDARMORSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGGROUNDARMORSLEVEL1
        },
        UpgradeId.ZERGGROUNDARMORSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGGROUNDARMORSLEVEL2
        },
        UpgradeId.ZERGMELEEWEAPONSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1
        },
        UpgradeId.ZERGMELEEWEAPONSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGMELEEWEAPONSLEVEL1
        },
        UpgradeId.ZERGMELEEWEAPONSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGMELEEWEAPONSLEVEL2
        },
        UpgradeId.ZERGMISSILEWEAPONSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1
        },
        UpgradeId.ZERGMISSILEWEAPONSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGMISSILEWEAPONSLEVEL1
        },
        UpgradeId.ZERGMISSILEWEAPONSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGMISSILEWEAPONSLEVEL2
        }
    },
    UnitTypeId.FACTORYTECHLAB: {
        UpgradeId.CYCLONELOCKONDAMAGEUPGRADE: {
            'ability': AbilityId.RESEARCH_CYCLONELOCKONDAMAGE
        },
        UpgradeId.DRILLCLAWS: {
            'ability': AbilityId.RESEARCH_DRILLINGCLAWS,
            'required_building': UnitTypeId.ARMORY
        },
        UpgradeId.HIGHCAPACITYBARRELS: {
            'ability': AbilityId.RESEARCH_INFERNALPREIGNITER
        },
        UpgradeId.SMARTSERVOS: {
            'ability': AbilityId.RESEARCH_SMARTSERVOS,
            'required_building': UnitTypeId.ARMORY
        }
    },
    UnitTypeId.FLEETBEACON: {
        UpgradeId.PHOENIXRANGEUPGRADE: {
            'ability': AbilityId.RESEARCH_PHOENIXANIONPULSECRYSTALS,
            'requires_power': True
        },
        UpgradeId.TEMPESTGROUNDATTACKUPGRADE: {
            'ability': AbilityId.FLEETBEACONRESEARCH_TEMPESTRESEARCHGROUNDATTACKUPGRADE,
            'requires_power': True
        },
        UpgradeId.VOIDRAYSPEEDUPGRADE: {
            'ability': AbilityId.FLEETBEACONRESEARCH_RESEARCHVOIDRAYSPEEDUPGRADE,
            'requires_power': True
        }
    },
    UnitTypeId.FORGE: {
        UpgradeId.PROTOSSGROUNDARMORSLEVEL1: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSGROUNDARMORSLEVEL2: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSGROUNDARMORSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSGROUNDARMORSLEVEL3: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSGROUNDARMORSLEVEL2,
            'requires_power': True
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL1: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL2: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSGROUNDWEAPONSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSGROUNDWEAPONSLEVEL3: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSGROUNDWEAPONSLEVEL2,
            'requires_power': True
        },
        UpgradeId.PROTOSSSHIELDSLEVEL1: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSSHIELDSLEVEL2: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSSHIELDSLEVEL1,
            'requires_power': True
        },
        UpgradeId.PROTOSSSHIELDSLEVEL3: {
            'ability': AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3,
            'required_building': UnitTypeId.TWILIGHTCOUNCIL,
            'required_upgrade': UpgradeId.PROTOSSSHIELDSLEVEL2,
            'requires_power': True
        }
    },
    UnitTypeId.FUSIONCORE: {
        UpgradeId.BATTLECRUISERENABLESPECIALIZATIONS: {
            'ability': AbilityId.RESEARCH_BATTLECRUISERWEAPONREFIT
        },
        UpgradeId.LIBERATORAGRANGEUPGRADE: {
            'ability': AbilityId.FUSIONCORERESEARCH_RESEARCHBALLISTICRANGE
        },
        UpgradeId.MEDIVACINCREASESPEEDBOOST: {
            'ability': AbilityId.FUSIONCORERESEARCH_RESEARCHRAPIDREIGNITIONSYSTEM
        }
    },
    UnitTypeId.GHOSTACADEMY: {
        UpgradeId.ENHANCEDSHOCKWAVES: {
            'ability': AbilityId.GHOSTACADEMYRESEARCH_RESEARCHENHANCEDSHOCKWAVES
        },
        UpgradeId.PERSONALCLOAKING: {
            'ability': AbilityId.RESEARCH_PERSONALCLOAKING
        }
    },
    UnitTypeId.GREATERSPIRE: {
        UpgradeId.ZERGFLYERARMORSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1
        },
        UpgradeId.ZERGFLYERARMORSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGFLYERARMORSLEVEL1
        },
        UpgradeId.ZERGFLYERARMORSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGFLYERARMORSLEVEL2
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGFLYERWEAPONSLEVEL1
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGFLYERWEAPONSLEVEL2
        }
    },
    UnitTypeId.HATCHERY: {
        UpgradeId.BURROW: {
            'ability': AbilityId.RESEARCH_BURROW
        },
        UpgradeId.OVERLORDSPEED: {
            'ability': AbilityId.RESEARCH_PNEUMATIZEDCARAPACE
        }
    },
    UnitTypeId.HIVE: {
        UpgradeId.BURROW: {
            'ability': AbilityId.RESEARCH_BURROW
        },
        UpgradeId.OVERLORDSPEED: {
            'ability': AbilityId.RESEARCH_PNEUMATIZEDCARAPACE
        }
    },
    UnitTypeId.HYDRALISKDEN: {
        UpgradeId.EVOLVEGROOVEDSPINES: {
            'ability': AbilityId.RESEARCH_GROOVEDSPINES
        },
        UpgradeId.EVOLVEMUSCULARAUGMENTS: {
            'ability': AbilityId.RESEARCH_MUSCULARAUGMENTS
        }
    },
    UnitTypeId.INFESTATIONPIT: {
        UpgradeId.INFESTORENERGYUPGRADE: {
            'ability': AbilityId.RESEARCH_PATHOGENGLANDS
        },
        UpgradeId.NEURALPARASITE: {
            'ability': AbilityId.RESEARCH_NEURALPARASITE
        }
    },
    UnitTypeId.LAIR: {
        UpgradeId.BURROW: {
            'ability': AbilityId.RESEARCH_BURROW
        },
        UpgradeId.OVERLORDSPEED: {
            'ability': AbilityId.RESEARCH_PNEUMATIZEDCARAPACE
        }
    },
    UnitTypeId.LURKERDENMP: {
        UpgradeId.DIGGINGCLAWS: {
            'ability': AbilityId.RESEARCH_ADAPTIVETALONS,
            'required_building': UnitTypeId.HIVE
        },
        UpgradeId.LURKERRANGE: {
            'ability': AbilityId.LURKERDENRESEARCH_RESEARCHLURKERRANGE,
            'required_building': UnitTypeId.HIVE
        }
    },
    UnitTypeId.ROACHWARREN: {
        UpgradeId.GLIALRECONSTITUTION: {
            'ability': AbilityId.RESEARCH_GLIALREGENERATION,
            'required_building': UnitTypeId.LAIR
        },
        UpgradeId.TUNNELINGCLAWS: {
            'ability': AbilityId.RESEARCH_TUNNELINGCLAWS,
            'required_building': UnitTypeId.LAIR
        }
    },
    UnitTypeId.ROBOTICSBAY: {
        UpgradeId.EXTENDEDTHERMALLANCE: {
            'ability': AbilityId.RESEARCH_EXTENDEDTHERMALLANCE,
            'requires_power': True
        },
        UpgradeId.GRAVITICDRIVE: {
            'ability': AbilityId.RESEARCH_GRAVITICDRIVE,
            'requires_power': True
        },
        UpgradeId.OBSERVERGRAVITICBOOSTER: {
            'ability': AbilityId.RESEARCH_GRAVITICBOOSTER,
            'requires_power': True
        }
    },
    UnitTypeId.SPAWNINGPOOL: {
        UpgradeId.ZERGLINGATTACKSPEED: {
            'ability': AbilityId.RESEARCH_ZERGLINGADRENALGLANDS,
            'required_building': UnitTypeId.HIVE
        },
        UpgradeId.ZERGLINGMOVEMENTSPEED: {
            'ability': AbilityId.RESEARCH_ZERGLINGMETABOLICBOOST
        }
    },
    UnitTypeId.SPIRE: {
        UpgradeId.ZERGFLYERARMORSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1
        },
        UpgradeId.ZERGFLYERARMORSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGFLYERARMORSLEVEL1
        },
        UpgradeId.ZERGFLYERARMORSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGFLYERARMORSLEVEL2
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL1: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL2: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2,
            'required_building': UnitTypeId.LAIR,
            'required_upgrade': UpgradeId.ZERGFLYERWEAPONSLEVEL1
        },
        UpgradeId.ZERGFLYERWEAPONSLEVEL3: {
            'ability': AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3,
            'required_building': UnitTypeId.HIVE,
            'required_upgrade': UpgradeId.ZERGFLYERWEAPONSLEVEL2
        }
    },
    UnitTypeId.STARPORTTECHLAB: {
        UpgradeId.BANSHEECLOAK: {
            'ability': AbilityId.RESEARCH_BANSHEECLOAKINGFIELD
        },
        UpgradeId.BANSHEESPEED: {
            'ability': AbilityId.RESEARCH_BANSHEEHYPERFLIGHTROTORS
        },
        UpgradeId.RAVENCORVIDREACTOR: {
            'ability': AbilityId.RESEARCH_RAVENCORVIDREACTOR
        }
    },
    UnitTypeId.TEMPLARARCHIVE: {
        UpgradeId.PSISTORMTECH: {
            'ability': AbilityId.RESEARCH_PSISTORM,
            'requires_power': True
        }
    },
    UnitTypeId.TWILIGHTCOUNCIL: {
        UpgradeId.ADEPTPIERCINGATTACK: {
            'ability': AbilityId.RESEARCH_ADEPTRESONATINGGLAIVES,
            'requires_power': True
        },
        UpgradeId.BLINKTECH: {
            'ability': AbilityId.RESEARCH_BLINK,
            'requires_power': True
        },
        UpgradeId.CHARGE: {
            'ability': AbilityId.RESEARCH_CHARGE,
            'requires_power': True
        }
    },
    UnitTypeId.ULTRALISKCAVERN: {
        UpgradeId.ANABOLICSYNTHESIS: {
            'ability': AbilityId.RESEARCH_ANABOLICSYNTHESIS
        },
        UpgradeId.CHITINOUSPLATING: {
            'ability': AbilityId.RESEARCH_CHITINOUSPLATING
        }
    }
}



from sc2.position import Point2


class PowerSource:

    @classmethod
    def from_proto(cls, proto):
        return cls(Point2.from_proto(proto.pos), proto.radius, proto.tag)

    def __init__(self, position, radius, unit_tag):
        assert isinstance(position, Point2)
        assert radius > 0
        self.position = position
        self.radius = radius
        self.unit_tag = unit_tag

    def covers(self, position):
        return self.position.distance_to(position) <= self.radius

    def __repr__(self):
        return f"PowerSource({self.position}, {self.radius})"


class PsionicMatrix:

    @classmethod
    def from_proto(cls, proto):
        return cls([PowerSource.from_proto(p) for p in proto])

    def __init__(self, sources):
        self.sources = sources

    def covers(self, position):
        return any(source.covers(position) for source in self.sources)



class ControlGroup(set):

    def __init__(self, units):
        super().__init__({unit.tag for unit in units})

    def __hash__(self):
        return hash(tuple(sorted(list(self))))

    def select_units(self, units):
        return units.filter(lambda unit: unit.tag in self)

    def missing_unit_tags(self, units):
        return {t for t in self if units.find_by_tag(t) is None}

    @property
    def amount(self) -> int:
        return len(self)

    @property
    def empty(self) -> bool:
        return not self

    def add_unit(self, unit):
        self.add(unit.tag)

    def add_units(self, units):
        for unit in units:
            self.add_unit(unit)

    def remove_unit(self, unit):
        self.discard(unit.tag)

    def remove_units(self, units):
        for unit in units:
            self.discard(unit.tag)



# THIS FILE WAS AUTOMATICALLY GENERATED BY "generate_dicts_from_data_json.py" DO NOT CHANGE MANUALLY!
# ANY CHANGE WILL BE OVERWRITTEN

from typing import Dict

from sc2.ids.ability_id import AbilityId

# from ..ids.buff_id import BuffId
# from ..ids.effect_id import EffectId

GENERIC_REDIRECT_ABILITIES: Dict[AbilityId, AbilityId] = {
    AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL1: AbilityId.RESEARCH_TERRANSHIPWEAPONS,
    AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL2: AbilityId.RESEARCH_TERRANSHIPWEAPONS,
    AbilityId.ARMORYRESEARCH_TERRANSHIPWEAPONSLEVEL3: AbilityId.RESEARCH_TERRANSHIPWEAPONS,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL1: AbilityId.RESEARCH_TERRANVEHICLEANDSHIPPLATING,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL2: AbilityId.RESEARCH_TERRANVEHICLEANDSHIPPLATING,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEANDSHIPPLATINGLEVEL3: AbilityId.RESEARCH_TERRANVEHICLEANDSHIPPLATING,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL1: AbilityId.RESEARCH_TERRANVEHICLEWEAPONS,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL2: AbilityId.RESEARCH_TERRANVEHICLEWEAPONS,
    AbilityId.ARMORYRESEARCH_TERRANVEHICLEWEAPONSLEVEL3: AbilityId.RESEARCH_TERRANVEHICLEWEAPONS,
    AbilityId.ATTACKPROTOSSBUILDING_ATTACKBUILDING: AbilityId.ATTACK,
    AbilityId.ATTACK_ATTACK: AbilityId.ATTACK,
    AbilityId.ATTACK_BATTLECRUISER: AbilityId.ATTACK,
    AbilityId.ATTACK_REDIRECT: AbilityId.ATTACK,
    AbilityId.BEHAVIOR_CLOAKOFF_BANSHEE: AbilityId.BEHAVIOR_CLOAKOFF,
    AbilityId.BEHAVIOR_CLOAKOFF_GHOST: AbilityId.BEHAVIOR_CLOAKOFF,
    AbilityId.BEHAVIOR_CLOAKON_BANSHEE: AbilityId.BEHAVIOR_CLOAKON,
    AbilityId.BEHAVIOR_CLOAKON_GHOST: AbilityId.BEHAVIOR_CLOAKON,
    AbilityId.BEHAVIOR_HOLDFIREOFF_GHOST: AbilityId.BEHAVIOR_HOLDFIREOFF,
    AbilityId.BEHAVIOR_HOLDFIREOFF_LURKER: AbilityId.BEHAVIOR_HOLDFIREOFF,
    AbilityId.BEHAVIOR_HOLDFIREON_GHOST: AbilityId.BEHAVIOR_HOLDFIREON,
    AbilityId.BEHAVIOR_HOLDFIREON_LURKER: AbilityId.BEHAVIOR_HOLDFIREON,
    AbilityId.BROODLORDQUEUE2_CANCEL: AbilityId.CANCEL_LAST,
    AbilityId.BROODLORDQUEUE2_CANCELSLOT: AbilityId.CANCEL_SLOT,
    AbilityId.BUILDINPROGRESSNYDUSCANAL_CANCEL: AbilityId.CANCEL,
    AbilityId.BUILDNYDUSCANAL_CANCEL: AbilityId.HALT,
    AbilityId.BUILD_CREEPTUMOR_QUEEN: AbilityId.BUILD_CREEPTUMOR,
    AbilityId.BUILD_CREEPTUMOR_TUMOR: AbilityId.BUILD_CREEPTUMOR,
    AbilityId.BUILD_REACTOR_BARRACKS: AbilityId.BUILD_REACTOR,
    AbilityId.BUILD_REACTOR_FACTORY: AbilityId.BUILD_REACTOR,
    AbilityId.BUILD_REACTOR_STARPORT: AbilityId.BUILD_REACTOR,
    AbilityId.BUILD_TECHLAB_BARRACKS: AbilityId.BUILD_TECHLAB,
    AbilityId.BUILD_TECHLAB_FACTORY: AbilityId.BUILD_TECHLAB,
    AbilityId.BUILD_TECHLAB_STARPORT: AbilityId.BUILD_TECHLAB,
    AbilityId.BURROWBANELINGDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWCREEPTUMORDOWN_BURROWDOWN: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_BANELING: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_DRONE: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_HYDRALISK: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_INFESTOR: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_INFESTORTERRAN: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_LURKER: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_QUEEN: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_RAVAGER: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_ROACH: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_SWARMHOST: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_ULTRALISK: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_WIDOWMINE: AbilityId.BURROWDOWN,
    AbilityId.BURROWDOWN_ZERGLING: AbilityId.BURROWDOWN,
    AbilityId.BURROWDRONEDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWHYDRALISKDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWINFESTORDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWLURKERMPDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWQUEENDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWRAVAGERDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWROACHDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.BURROWUP_BANELING: AbilityId.BURROWUP,
    AbilityId.BURROWUP_DRONE: AbilityId.BURROWUP,
    AbilityId.BURROWUP_HYDRALISK: AbilityId.BURROWUP,
    AbilityId.BURROWUP_INFESTOR: AbilityId.BURROWUP,
    AbilityId.BURROWUP_INFESTORTERRAN: AbilityId.BURROWUP,
    AbilityId.BURROWUP_LURKER: AbilityId.BURROWUP,
    AbilityId.BURROWUP_QUEEN: AbilityId.BURROWUP,
    AbilityId.BURROWUP_RAVAGER: AbilityId.BURROWUP,
    AbilityId.BURROWUP_ROACH: AbilityId.BURROWUP,
    AbilityId.BURROWUP_SWARMHOST: AbilityId.BURROWUP,
    AbilityId.BURROWUP_ULTRALISK: AbilityId.BURROWUP,
    AbilityId.BURROWUP_WIDOWMINE: AbilityId.BURROWUP,
    AbilityId.BURROWUP_ZERGLING: AbilityId.BURROWUP,
    AbilityId.BURROWZERGLINGDOWN_CANCEL: AbilityId.CANCEL,
    AbilityId.CANCELSLOT_ADDON: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_HANGARQUEUE5: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_QUEUE1: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_QUEUE5: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_QUEUECANCELTOSELECTION: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_QUEUEPASSIVE: AbilityId.CANCEL_SLOT,
    AbilityId.CANCELSLOT_QUEUEPASSIVECANCELTOSELECTION: AbilityId.CANCEL_SLOT,
    AbilityId.CANCEL_ADEPTPHASESHIFT: AbilityId.CANCEL,
    AbilityId.CANCEL_ADEPTSHADEPHASESHIFT: AbilityId.CANCEL,
    AbilityId.CANCEL_BARRACKSADDON: AbilityId.CANCEL,
    AbilityId.CANCEL_BUILDINPROGRESS: AbilityId.CANCEL,
    AbilityId.CANCEL_CREEPTUMOR: AbilityId.CANCEL,
    AbilityId.CANCEL_FACTORYADDON: AbilityId.CANCEL,
    AbilityId.CANCEL_GRAVITONBEAM: AbilityId.CANCEL,
    AbilityId.CANCEL_HANGARQUEUE5: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_LOCKON: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHBROODLORD: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHGREATERSPIRE: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHHIVE: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHLAIR: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHLURKER: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHMOTHERSHIP: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHORBITAL: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHOVERLORDTRANSPORT: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHOVERSEER: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHPLANETARYFORTRESS: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHRAVAGER: AbilityId.CANCEL,
    AbilityId.CANCEL_MORPHTHOREXPLOSIVEMODE: AbilityId.CANCEL,
    AbilityId.CANCEL_NEURALPARASITE: AbilityId.CANCEL,
    AbilityId.CANCEL_NUKE: AbilityId.CANCEL,
    AbilityId.CANCEL_QUEUE1: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_QUEUE5: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_QUEUEADDON: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_QUEUECANCELTOSELECTION: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_QUEUEPASIVE: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_QUEUEPASSIVECANCELTOSELECTION: AbilityId.CANCEL_LAST,
    AbilityId.CANCEL_SPINECRAWLERROOT: AbilityId.CANCEL,
    AbilityId.CANCEL_SPORECRAWLERROOT: AbilityId.CANCEL,
    AbilityId.CANCEL_STARPORTADDON: AbilityId.CANCEL,
    AbilityId.CANCEL_STASISTRAP: AbilityId.CANCEL,
    AbilityId.CANCEL_VOIDRAYPRISMATICALIGNMENT: AbilityId.CANCEL,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL1: AbilityId.RESEARCH_PROTOSSAIRARMOR,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL2: AbilityId.RESEARCH_PROTOSSAIRARMOR,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRARMORLEVEL3: AbilityId.RESEARCH_PROTOSSAIRARMOR,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL1: AbilityId.RESEARCH_PROTOSSAIRWEAPONS,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL2: AbilityId.RESEARCH_PROTOSSAIRWEAPONS,
    AbilityId.CYBERNETICSCORERESEARCH_PROTOSSAIRWEAPONSLEVEL3: AbilityId.RESEARCH_PROTOSSAIRWEAPONS,
    AbilityId.DEFILERMPBURROW_BURROWDOWN: AbilityId.BURROWDOWN,
    AbilityId.DEFILERMPBURROW_CANCEL: AbilityId.CANCEL,
    AbilityId.DEFILERMPUNBURROW_BURROWUP: AbilityId.BURROWUP,
    AbilityId.EFFECT_BLINK_STALKER: AbilityId.EFFECT_BLINK,
    AbilityId.EFFECT_MASSRECALL_MOTHERSHIPCORE: AbilityId.EFFECT_MASSRECALL,
    AbilityId.EFFECT_MASSRECALL_NEXUS: AbilityId.EFFECT_MASSRECALL,
    AbilityId.EFFECT_MASSRECALL_STRATEGICRECALL: AbilityId.EFFECT_MASSRECALL,
    AbilityId.EFFECT_REPAIR_MULE: AbilityId.EFFECT_REPAIR,
    AbilityId.EFFECT_REPAIR_REPAIRDRONE: AbilityId.EFFECT_REPAIR,
    AbilityId.EFFECT_REPAIR_SCV: AbilityId.EFFECT_REPAIR,
    AbilityId.EFFECT_SHADOWSTRIDE: AbilityId.EFFECT_BLINK,
    AbilityId.EFFECT_SPRAY_PROTOSS: AbilityId.EFFECT_SPRAY,
    AbilityId.EFFECT_SPRAY_TERRAN: AbilityId.EFFECT_SPRAY,
    AbilityId.EFFECT_SPRAY_ZERG: AbilityId.EFFECT_SPRAY,
    AbilityId.EFFECT_STIM_MARAUDER: AbilityId.EFFECT_STIM,
    AbilityId.EFFECT_STIM_MARAUDER_REDIRECT: AbilityId.EFFECT_STIM,
    AbilityId.EFFECT_STIM_MARINE: AbilityId.EFFECT_STIM,
    AbilityId.EFFECT_STIM_MARINE_REDIRECT: AbilityId.EFFECT_STIM,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL1: AbilityId.RESEARCH_TERRANINFANTRYARMOR,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL2: AbilityId.RESEARCH_TERRANINFANTRYARMOR,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYARMORLEVEL3: AbilityId.RESEARCH_TERRANINFANTRYARMOR,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL1: AbilityId.RESEARCH_TERRANINFANTRYWEAPONS,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL2: AbilityId.RESEARCH_TERRANINFANTRYWEAPONS,
    AbilityId.ENGINEERINGBAYRESEARCH_TERRANINFANTRYWEAPONSLEVEL3: AbilityId.RESEARCH_TERRANINFANTRYWEAPONS,
    AbilityId.FORCEFIELD_CANCEL: AbilityId.CANCEL,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL1: AbilityId.RESEARCH_PROTOSSGROUNDARMOR,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL2: AbilityId.RESEARCH_PROTOSSGROUNDARMOR,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDARMORLEVEL3: AbilityId.RESEARCH_PROTOSSGROUNDARMOR,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL1: AbilityId.RESEARCH_PROTOSSGROUNDWEAPONS,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL2: AbilityId.RESEARCH_PROTOSSGROUNDWEAPONS,
    AbilityId.FORGERESEARCH_PROTOSSGROUNDWEAPONSLEVEL3: AbilityId.RESEARCH_PROTOSSGROUNDWEAPONS,
    AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL1: AbilityId.RESEARCH_PROTOSSSHIELDS,
    AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL2: AbilityId.RESEARCH_PROTOSSSHIELDS,
    AbilityId.FORGERESEARCH_PROTOSSSHIELDSLEVEL3: AbilityId.RESEARCH_PROTOSSSHIELDS,
    AbilityId.HALT_BUILDING: AbilityId.HALT,
    AbilityId.HALT_TERRANBUILD: AbilityId.HALT,
    AbilityId.HARVEST_GATHER_DRONE: AbilityId.HARVEST_GATHER,
    AbilityId.HARVEST_GATHER_MULE: AbilityId.HARVEST_GATHER,
    AbilityId.HARVEST_GATHER_PROBE: AbilityId.HARVEST_GATHER,
    AbilityId.HARVEST_GATHER_SCV: AbilityId.HARVEST_GATHER,
    AbilityId.HARVEST_RETURN_DRONE: AbilityId.HARVEST_RETURN,
    AbilityId.HARVEST_RETURN_MULE: AbilityId.HARVEST_RETURN,
    AbilityId.HARVEST_RETURN_PROBE: AbilityId.HARVEST_RETURN,
    AbilityId.HARVEST_RETURN_SCV: AbilityId.HARVEST_RETURN,
    AbilityId.HOLDPOSITION_BATTLECRUISER: AbilityId.HOLDPOSITION,
    AbilityId.HOLDPOSITION_HOLD: AbilityId.HOLDPOSITION,
    AbilityId.LAND_BARRACKS: AbilityId.LAND,
    AbilityId.LAND_COMMANDCENTER: AbilityId.LAND,
    AbilityId.LAND_FACTORY: AbilityId.LAND,
    AbilityId.LAND_ORBITALCOMMAND: AbilityId.LAND,
    AbilityId.LAND_STARPORT: AbilityId.LAND,
    AbilityId.LIFT_BARRACKS: AbilityId.LIFT,
    AbilityId.LIFT_COMMANDCENTER: AbilityId.LIFT,
    AbilityId.LIFT_FACTORY: AbilityId.LIFT,
    AbilityId.LIFT_ORBITALCOMMAND: AbilityId.LIFT,
    AbilityId.LIFT_STARPORT: AbilityId.LIFT,
    AbilityId.LOADALL_COMMANDCENTER: AbilityId.LOADALL,
    AbilityId.LOAD_BUNKER: AbilityId.LOAD,
    AbilityId.LOAD_MEDIVAC: AbilityId.LOAD,
    AbilityId.LOAD_NYDUSNETWORK: AbilityId.LOAD,
    AbilityId.LOAD_NYDUSWORM: AbilityId.LOAD,
    AbilityId.LOAD_OVERLORD: AbilityId.LOAD,
    AbilityId.LOAD_WARPPRISM: AbilityId.LOAD,
    AbilityId.MERGEABLE_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHBACKTOGATEWAY_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEPURIFIERTOWERDEBRIS_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEROCKTOWERDEBRISRAMPLEFTGREEN_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEROCKTOWERDEBRISRAMPLEFT_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEROCKTOWERDEBRISRAMPRIGHTGREEN_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEROCKTOWERDEBRISRAMPRIGHT_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLEROCKTOWERDEBRIS_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLETERRANTOWERDEBRISRAMPLEFT_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLETERRANTOWERDEBRISRAMPRIGHT_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOCOLLAPSIBLETERRANTOWERDEBRIS_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTODEVOURERMP_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOGUARDIANMP_CANCEL: AbilityId.CANCEL,
    AbilityId.MORPHTOSWARMHOSTBURROWEDMP_CANCEL: AbilityId.CANCEL,
    AbilityId.MOVE_BATTLECRUISER: AbilityId.MOVE,
    AbilityId.MOVE_MOVE: AbilityId.MOVE,
    AbilityId.PATROL_BATTLECRUISER: AbilityId.PATROL,
    AbilityId.PATROL_PATROL: AbilityId.PATROL,
    AbilityId.PHASINGMODE_CANCEL: AbilityId.CANCEL,
    AbilityId.PROTOSSBUILD_CANCEL: AbilityId.HALT,
    AbilityId.QUEENBUILD_CANCEL: AbilityId.HALT,
    AbilityId.RALLY_BUILDING: AbilityId.RALLY_UNITS,
    AbilityId.RALLY_COMMANDCENTER: AbilityId.RALLY_WORKERS,
    AbilityId.RALLY_HATCHERY_UNITS: AbilityId.RALLY_UNITS,
    AbilityId.RALLY_HATCHERY_WORKERS: AbilityId.RALLY_WORKERS,
    AbilityId.RALLY_MORPHING_UNIT: AbilityId.RALLY_UNITS,
    AbilityId.RALLY_NEXUS: AbilityId.RALLY_WORKERS,
    AbilityId.RESEARCH_ZERGFLYERARMORLEVEL1: AbilityId.RESEARCH_ZERGFLYERARMOR,
    AbilityId.RESEARCH_ZERGFLYERARMORLEVEL2: AbilityId.RESEARCH_ZERGFLYERARMOR,
    AbilityId.RESEARCH_ZERGFLYERARMORLEVEL3: AbilityId.RESEARCH_ZERGFLYERARMOR,
    AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL1: AbilityId.RESEARCH_ZERGFLYERATTACK,
    AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL2: AbilityId.RESEARCH_ZERGFLYERATTACK,
    AbilityId.RESEARCH_ZERGFLYERATTACKLEVEL3: AbilityId.RESEARCH_ZERGFLYERATTACK,
    AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL1: AbilityId.RESEARCH_ZERGGROUNDARMOR,
    AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL2: AbilityId.RESEARCH_ZERGGROUNDARMOR,
    AbilityId.RESEARCH_ZERGGROUNDARMORLEVEL3: AbilityId.RESEARCH_ZERGGROUNDARMOR,
    AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL1: AbilityId.RESEARCH_ZERGMELEEWEAPONS,
    AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL2: AbilityId.RESEARCH_ZERGMELEEWEAPONS,
    AbilityId.RESEARCH_ZERGMELEEWEAPONSLEVEL3: AbilityId.RESEARCH_ZERGMELEEWEAPONS,
    AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL1: AbilityId.RESEARCH_ZERGMISSILEWEAPONS,
    AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL2: AbilityId.RESEARCH_ZERGMISSILEWEAPONS,
    AbilityId.RESEARCH_ZERGMISSILEWEAPONSLEVEL3: AbilityId.RESEARCH_ZERGMISSILEWEAPONS,
    AbilityId.SCAN_MOVE: AbilityId.ATTACK,
    AbilityId.SHIELDBATTERYRECHARGEEX5_STOP: AbilityId.CANCEL,
    AbilityId.SPINECRAWLERROOT_SPINECRAWLERROOT: AbilityId.MORPH_ROOT,
    AbilityId.SPINECRAWLERUPROOT_CANCEL: AbilityId.CANCEL,
    AbilityId.SPINECRAWLERUPROOT_SPINECRAWLERUPROOT: AbilityId.MORPH_UPROOT,
    AbilityId.SPORECRAWLERROOT_SPORECRAWLERROOT: AbilityId.MORPH_ROOT,
    AbilityId.SPORECRAWLERUPROOT_CANCEL: AbilityId.CANCEL,
    AbilityId.SPORECRAWLERUPROOT_SPORECRAWLERUPROOT: AbilityId.MORPH_UPROOT,
    AbilityId.STOP_BATTLECRUISER: AbilityId.STOP,
    AbilityId.STOP_BUILDING: AbilityId.STOP,
    AbilityId.STOP_CHEER: AbilityId.STOP,
    AbilityId.STOP_DANCE: AbilityId.STOP,
    AbilityId.STOP_HOLDFIRESPECIAL: AbilityId.STOP,
    AbilityId.STOP_REDIRECT: AbilityId.STOP,
    AbilityId.STOP_STOP: AbilityId.STOP,
    AbilityId.TESTZERG_CANCEL: AbilityId.CANCEL,
    AbilityId.THORAPMODE_CANCEL: AbilityId.CANCEL,
    AbilityId.TRANSPORTMODE_CANCEL: AbilityId.CANCEL,
    AbilityId.UNLOADALLAT_MEDIVAC: AbilityId.UNLOADALLAT,
    AbilityId.UNLOADALLAT_OVERLORD: AbilityId.UNLOADALLAT,
    AbilityId.UNLOADALLAT_WARPPRISM: AbilityId.UNLOADALLAT,
    AbilityId.UNLOADALL_BUNKER: AbilityId.UNLOADALL,
    AbilityId.UNLOADALL_COMMANDCENTER: AbilityId.UNLOADALL,
    AbilityId.UNLOADALL_NYDASNETWORK: AbilityId.UNLOADALL,
    AbilityId.UNLOADALL_NYDUSWORM: AbilityId.UNLOADALL,
    AbilityId.UNLOADALL_WARPPRISM: AbilityId.UNLOADALL,
    AbilityId.UNLOADUNIT_BUNKER: AbilityId.UNLOADUNIT,
    AbilityId.UNLOADUNIT_COMMANDCENTER: AbilityId.UNLOADUNIT,
    AbilityId.UNLOADUNIT_MEDIVAC: AbilityId.UNLOADUNIT,
    AbilityId.UNLOADUNIT_NYDASNETWORK: AbilityId.UNLOADUNIT,
    AbilityId.UNLOADUNIT_OVERLORD: AbilityId.UNLOADUNIT,
    AbilityId.UNLOADUNIT_WARPPRISM: AbilityId.UNLOADUNIT,
    AbilityId.UPGRADETOWARPGATE_CANCEL: AbilityId.CANCEL,
    AbilityId.WARPABLE_CANCEL: AbilityId.CANCEL,
    AbilityId.WIDOWMINEBURROW_CANCEL: AbilityId.CANCEL,
    AbilityId.ZERGBUILD_CANCEL: AbilityId.HALT
}



from functools import wraps


def property_cache_forever(f):

    @wraps(f)
    def inner(self):
        property_cache = "_cache_" + f.__name__
        cache_updated = hasattr(self, property_cache)
        if not cache_updated:
            setattr(self, property_cache, f(self))
        cache = getattr(self, property_cache)
        return cache

    return property(inner)


def property_cache_once_per_frame(f):
    """This decorator caches the return value for one game loop,
    then clears it if it is accessed in a different game loop.
    Only works on properties of the bot object, because it requires
    access to self.state.game_loop"""

    @wraps(f)
    def inner(self):
        property_cache = "_cache_" + f.__name__
        state_cache = "_frame_" + f.__name__
        cache_updated = getattr(self, state_cache, -1) == self.state.game_loop
        if not cache_updated:
            setattr(self, property_cache, f(self))
            setattr(self, state_cache, self.state.game_loop)

        cache = getattr(self, property_cache)
        should_copy = callable(getattr(cache, "copy", None))
        if should_copy:
            return cache.copy()
        return cache

    return property(inner)


def property_cache_once_per_frame_no_copy(f):
    """This decorator caches the return value for one game loop,
    then clears it if it is accessed in a different game loop.
    Only works on properties of the bot object, because it requires
    access to self.state.game_loop

    This decorator compared to the above runs a little faster, however you should only use this decorator if you are sure that you do not modify the mutable once it is calculated and cached."""

    @wraps(f)
    def inner(self):
        property_cache = "_cache_" + f.__name__
        state_cache = "_frame_" + f.__name__
        cache_updated = getattr(self, state_cache, -1) == self.state.game_loop
        if not cache_updated:
            setattr(self, property_cache, f(self))
            setattr(self, state_cache, self.state.game_loop)

        cache = getattr(self, property_cache)
        return cache

    return property(inner)


def property_immutable_cache(f):
    """ This cache should only be used on properties that return an immutable object (bool, str, int, float, tuple, Unit, Point2, Point3) """

    @wraps(f)
    def inner(self):
        if f.__name__ not in self.cache:
            self.cache[f.__name__] = f(self)
        return self.cache[f.__name__]

    return property(inner)


def property_mutable_cache(f):
    """ This cache should only be used on properties that return a mutable object (Units, list, set, dict, Counter) """

    @wraps(f)
    def inner(self):
        if f.__name__ not in self.cache:
            self.cache[f.__name__] = f(self)
        return self.cache[f.__name__].copy()

    return property(inner)



import math
import warnings
from math import pow

import numpy as np

from sc2.game_state import GameState
from sc2.unit import Unit
from sc2.units import Units

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from scipy.spatial.distance import pdist, cdist

from typing import Generator, Iterable, Tuple


class DistanceCalculation:

    def __init__(self):
        self.state: GameState = None
        # self._generated_frame = -100
        self._generated_frame2 = -100
        # Pdist condensed vector generated by scipy pdist, half the size of the cdist matrix as 1d array
        self._cached_pdist: np.ndarray = None
        self._cached_cdist: np.ndarray = None

    @property
    def _units_count(self) -> int:
        return len(self.all_units)

    @property
    def _pdist(self) -> np.ndarray:
        """ As property, so it will be recalculated each time it is called, or return from cache if it is called multiple times in teh same game_loop. """
        if self._generated_frame2 != self.state.game_loop:
            return self.calculate_distances()
        return self._cached_pdist

    @property
    def _cdist(self) -> np.ndarray:
        """ As property, so it will be recalculated each time it is called, or return from cache if it is called multiple times in teh same game_loop. """
        if self._generated_frame2 != self.state.game_loop:
            return self.calculate_distances()
        return self._cached_cdist

    def _calculate_distances_method1(self) -> np.ndarray:
        self._generated_frame2 = self.state.game_loop
        # Converts tuple [(1, 2), (3, 4)] to flat list like [1, 2, 3, 4]
        flat_positions = (coord for unit in self.all_units for coord in unit.position_tuple)
        # Converts to numpy array, then converts the flat array back to shape (n, 2): [[1, 2], [3, 4]]
        positions_array: np.ndarray = np.fromiter(flat_positions, dtype=np.float,
                                                  count=2 * self._units_count).reshape((self._units_count, 2))
        assert len(positions_array) == self._units_count
        # See performance benchmarks
        self._cached_pdist = pdist(positions_array, "sqeuclidean")

        return self._cached_pdist

    def _calculate_distances_method2(self) -> np.ndarray:
        self._generated_frame2 = self.state.game_loop
        # Converts tuple [(1, 2), (3, 4)] to flat list like [1, 2, 3, 4]
        flat_positions = (coord for unit in self.all_units for coord in unit.position_tuple)
        # Converts to numpy array, then converts the flat array back to shape (n, 2): [[1, 2], [3, 4]]
        positions_array: np.ndarray = np.fromiter(flat_positions, dtype=np.float,
                                                  count=2 * self._units_count).reshape((self._units_count, 2))
        assert len(positions_array) == self._units_count
        # See performance benchmarks
        self._cached_cdist = cdist(positions_array, positions_array, "sqeuclidean")

        return self._cached_cdist

    def _calculate_distances_method3(self) -> np.ndarray:
        """ Nearly same as above, but without asserts"""
        self._generated_frame2 = self.state.game_loop
        flat_positions = (coord for unit in self.all_units for coord in unit.position_tuple)
        positions_array: np.ndarray = np.fromiter(flat_positions, dtype=np.float,
                                                  count=2 * self._units_count).reshape((-1, 2))
        # See performance benchmarks
        self._cached_cdist = cdist(positions_array, positions_array, "sqeuclidean")

        return self._cached_cdist

    # Helper functions

    def square_to_condensed(self, i, j) -> int:
        # Converts indices of a square matrix to condensed matrix
        # https://stackoverflow.com/a/36867493/10882657
        assert i != j, "No diagonal elements in condensed matrix! Diagonal elements are zero"
        if i < j:
            i, j = j, i
        return self._units_count * j - j * (j + 1) // 2 + i - 1 - j

    def convert_tuple_to_numpy_array(self, pos: Tuple[float, float]) -> np.ndarray:
        """ Converts a single position to a 2d numpy array with 1 row and 2 columns. """
        return np.fromiter(pos, dtype=float, count=2).reshape((1, 2))

    # Fast and simple calculation functions

    def distance_math_hypot(self, p1: Tuple[float, float], p2: Tuple[float, float]):
        return math.hypot(p1[0] - p2[0], p1[1] - p2[1])

    def distance_math_hypot_squared(self, p1: Tuple[float, float], p2: Tuple[float, float]):
        return pow(p1[0] - p2[0], 2) + pow(p1[1] - p2[1], 2)

    def _distance_squared_unit_to_unit_method0(self, unit1: Unit, unit2: Unit) -> float:
        return self.distance_math_hypot_squared(unit1.position_tuple, unit2.position_tuple)

    # Distance calculation using the pre-calculated matrix above

    def _distance_squared_unit_to_unit_method1(self, unit1: Unit, unit2: Unit) -> float:
        # If checked on units if they have the same tag, return distance 0 as these are not in the 1 dimensional pdist array - would result in an error otherwise
        if unit1.tag == unit2.tag:
            return 0
        # Calculate index, needs to be after pdist has been calculated and cached
        condensed_index = self.square_to_condensed(unit1.distance_calculation_index, unit2.distance_calculation_index)
        assert condensed_index < len(
            self._cached_pdist
        ), f"Condensed index is larger than amount of calculated distances: {condensed_index} < {len(self._cached_pdist)}, units that caused the assert error: {unit1} and {unit2}"
        distance = self._pdist[condensed_index]
        return distance

    def _distance_squared_unit_to_unit_method2(self, unit1: Unit, unit2: Unit) -> float:
        # Calculate index, needs to be after cdist has been calculated and cached
        return self._cdist[unit1.distance_calculation_index, unit2.distance_calculation_index]

    # Distance calculation using the fastest distance calculation functions

    def _distance_pos_to_pos(self, pos1: Tuple[float, float], pos2: Tuple[float, float]) -> float:
        return self.distance_math_hypot(pos1, pos2)

    def _distance_units_to_pos(self, units: Units, pos: Tuple[float, float]) -> Generator[float, None, None]:
        """ This function does not scale well, if len(units) > 100 it gets fairly slow """
        return (self.distance_math_hypot(u.position_tuple, pos) for u in units)

    def _distance_unit_to_points(self, unit: Unit, points: Iterable[Tuple[float,
                                                                          float]]) -> Generator[float, None, None]:
        """ This function does not scale well, if len(points) > 100 it gets fairly slow """
        pos = unit.position_tuple
        return (self.distance_math_hypot(p, pos) for p in points)

    def _distances_override_functions(self, method: int = 0):
        """Overrides the internal distance calculation functions at game start in bot_ai.py self._prepare_start() function
        method 0: Use python's math.hypot
        The following methods calculate the distances between all units once:
        method 1: Use scipy's pdist condensed matrix (1d array)
        method 2: Use scipy's cidst square matrix (2d array)
        method 3: Use scipy's cidst square matrix (2d array) without asserts (careful: very weird error messages, but maybe slightly faster)"""
        assert 0 <= method <= 3, f"Selected method was: {method}"
        if method == 0:
            self._distance_squared_unit_to_unit = self._distance_squared_unit_to_unit_method0
        elif method == 1:
            self._distance_squared_unit_to_unit = self._distance_squared_unit_to_unit_method1
            self.calculate_distances = self._calculate_distances_method1
        elif method == 2:
            self._distance_squared_unit_to_unit = self._distance_squared_unit_to_unit_method2
            self.calculate_distances = self._calculate_distances_method2
        elif method == 3:
            self._distance_squared_unit_to_unit = self._distance_squared_unit_to_unit_method2
            self.calculate_distances = self._calculate_distances_method3
